@@ -16,6 +16,7 @@
  * express or implied warranty.
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -97,6 +98,20 @@ int et_list_unlock(void)
 	return 0;
 }
 
+typedef char *(*gettextf) (const char *);
+
+static gettextf com_err_gettext = NULL;
+
+gettextf set_com_err_gettext(gettextf new_proc)
+{
+    gettextf x = com_err_gettext;
+
+    com_err_gettext = new_proc;
+
+    return x;
+}
+
+
 const char * error_message (errcode_t code)
 {
     int offset;
@@ -130,7 +145,10 @@ const char * error_message (errcode_t code)
 	    } else {
 		const char *msg = et->table->msgs[offset];
 		et_list_unlock();
-		return msg;
+		if (com_err_gettext)
+		    return (*com_err_gettext)(msg);
+		else
+		    return msg;
 	    }
 	}
     }
@@ -142,7 +160,10 @@ const char * error_message (errcode_t code)
 	    } else {
 		const char *msg = et->table->msgs[offset];
 		et_list_unlock();
-		return msg;
+		if (com_err_gettext)
+		    return (*com_err_gettext)(msg);
+		else
+		    return msg;
 	    }
 	}
     }
@@ -187,7 +208,9 @@ static char *safe_getenv(const char *arg)
 #endif
 #endif
 
-#ifdef HAVE___SECURE_GETENV
+#if defined(HAVE_SECURE_GETENV)
+	return secure_getenv(arg);
+#elif defined(HAVE___SECURE_GETENV)
 	return __secure_getenv(arg);
 #else
 	return getenv(arg);
