@@ -440,6 +440,10 @@ int main (int argc, char ** argv)
 		    !print_min_size)
 			checkit = 1;
 
+		if ((fs->super->s_free_blocks_count > fs->super->s_blocks_count) ||
+		    (fs->super->s_free_inodes_count > fs->super->s_inodes_count))
+			checkit = 1;
+
 		if (checkit) {
 			fprintf(stderr,
 				_("Please run 'e2fsck -f %s' first.\n\n"),
@@ -505,7 +509,7 @@ int main (int argc, char ** argv)
 		new_size = max_size;
 		/* Round down to an even multiple of a pagesize */
 		if (sys_page_size > blocksize)
-			new_size &= ~((sys_page_size / blocksize)-1);
+			new_size &= ~((blk64_t)((sys_page_size / blocksize)-1));
 	}
 	/* If changing 64bit, don't change the filesystem size. */
 	if (flags & (RESIZE_DISABLE_64BIT | RESIZE_ENABLE_64BIT)) {
